from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Self

from pydantic import BaseModel, NonNegativeInt as uint, PrivateAttr, ValidationError, model_validator

from mortis.aff.events.arc import Arc, ScaledArctap
from mortis.aff.events.camera import Camera
from mortis.aff.events.hold import Hold, HoldFloat
from mortis.aff.events.scenecontrol import SceneControl
from mortis.aff.events.tap import Tap, TapFloat
from mortis.aff.events.timing import Timing
from mortis.aff.types.models import AFFEvent
from mortis.aff.lexer.analyse import analyse_command, analyse_timinggroup_footer, analyse_timinggroup_header
from mortis.aff.lexer.token import tokenize
from mortis.globcfg import GlobalConfig
from mortis.utils import get_default_model_cfg


__all__ = ['parse_event', 'TimingGroup']

__registered_aff_events__: list[type[AFFEvent]] = []

def get_registered_aff_events():
	global __registered_aff_events__
	if not __registered_aff_events__:
		__registered_aff_events__ = [Timing, Arc, Tap, TapFloat, Hold, HoldFloat, SceneControl, Camera]
		if GlobalConfig.uses_scaled_arctap:
			__registered_aff_events__.insert(1, ScaledArctap)
	
	return __registered_aff_events__


def parse_event(
	command: str,
	args: tuple,
	extras: tuple | None = None,
) -> AFFEvent:
	
	errors: dict[str, Exception] = {}
	registered = get_registered_aff_events()
	for eventcls in registered:
		if command in eventcls.commands:
			try:
				return eventcls.from_args(args, extras)
			except Exception as e:
				errors[eventcls.__name__] = e
	
	input = {'command': command, 'args': args, 'extras': extras}
	raise ValidationError.from_exception_data(
		title = f'parsing AFF event from {input}',
		line_errors = [{
			'type': 'value_error',
			'loc': ('parse_aff_event', ),
			'input': input,
			'ctx': {'error': errors},
		}]
	)

__timing_group_attrs__ = ('noinput', 'fadingholds', 'anglex', 'angley')
class TimingGroup(BaseModel):

	model_config = get_default_model_cfg()

	noinput: bool | None = None
	fadingholds: bool | None = None
	anglex: uint | None = None
	angley: uint | None = None

	_events: list[AFFEvent] = PrivateAttr(default_factory=list)


	@property
	def args(self) -> dict[str, bool | uint | None]:
		return {
			name: current
			for name in __timing_group_attrs__
			if (current := getattr(self, name)) is not None
		}
	
	@args.setter
	def args(self, args: Any) -> None:
		if isinstance(args, dict):
			valid_args = self.load_args_from_dict(args)
		elif isinstance(args, str):
			valid_args = self.load_args_from_str(args)
		elif args is None:
			del self.args
			return
		else:
			raise TypeError(f'Invalid type for TimingGroup args: {type(args).__name__!r}')
		
		for name, value in valid_args.items():
			setattr(self, name, value)

	@args.deleter
	def args(self) -> None:
		for name in __timing_group_attrs__:
			setattr(self, name, None)
	

	@classmethod
	def load_args_from_dict(cls, args: dict[str, Any]) -> dict[str, bool | uint | None]:
		valid_args = {}
		for name in args:
			if name in __timing_group_attrs__:
				valid_args[name] = args[name]
			else:
				raise ValueError(f'Unknown TimingGroup argument: {name!r}')
		
		return valid_args
	
	@classmethod
	def load_args_from_str(cls, args_str: str) -> dict[str, bool | uint | None]:
		args: dict[str, bool | uint | None] = {}
		segments: list[str] = [s.strip() for s in args_str.split(',')]

		for seg in segments:
			if seg == 'noinput':
				args['noinput'] = True

			elif seg == 'fadingholds':
				args['fadingholds'] = True

			elif seg.startswith('anglex'):
				value_str = seg[len('anglex'):].strip()
				args['anglex'] = int(value_str)

			elif seg.startswith('angley'):
				value_str = seg[len('angley'):].strip()
				args['angley'] = int(value_str)
			
			else:
				raise ValueError(f'Unknown TimingGroup argument: {seg!r}')
		
		return args
	

	def dump_args_to_dict(self) -> dict:
		return {
			attr: current
			for attr in __timing_group_attrs__
			if (current := getattr(self, attr)) is not None
		}

	def dump_args_to_str(self) -> str:		
		segments = []
		for attr, current in self.dump_args_to_dict().items():
			if current is True:
				segments.append(attr)
			else:
				segments.append(f'{attr}{current}')
		return '_'.join(segments)

	
	def iter_events(self) -> Iterator[AFFEvent]:
		yield from self._events

	@property
	def events(self) -> tuple[AFFEvent, ...]:
		return tuple(self._events)

	def find_event(self, event: AFFEvent) -> int:
		for i, ev in enumerate(self._events):
			if ev is event:
				return i
		return -1

	def __contains__(self, event: AFFEvent) -> bool:
		return self.find_event(event) != -1

	def add_event(self, event: AFFEvent) -> None:
		if not isinstance(event, AFFEvent):
			raise TypeError(f'Event must be of type {AFFEvent.__name__}')
		
		if event in self:
			raise ValueError('Event is already in this group')
		self._events.append(event)
	
	def remove_event(self, event: AFFEvent) -> None:
		idx = self.find_event(event)
		if idx == -1:
			raise ValueError('Event is not found in this group')
		del self._events[idx]
	
	def clear_events(self) -> None:
		self._events = []

	@property
	def required_hitsounds(self) -> set[str]:
		hitsounds = set()
		for event in self.events:
			if not isinstance(event, Arc):
				continue
			hitsound = event.unwrap_hitsound()
			if hitsound is not None:
				hitsounds.add(hitsound)
		return hitsounds
	
	@classmethod
	def from_str(cls, contents: str, *, events_only: bool=False) -> Self:
		instance = cls()

		lines = contents.split('\n')
		tokens_list = list(map(tokenize, lines))
		if events_only:
			evtoks_list = tokens_list

		else:
			header_toks = tokens_list[0]
			evtoks_list = tokens_list[1:-1]
			footer_toks = tokens_list[-1]

			header_args = analyse_timinggroup_header(header_toks)
			analyse_timinggroup_footer(footer_toks)
			cls.args = header_args
		
		for evtoks in evtoks_list:
			command, args, extras = analyse_command(evtoks)
			event = parse_event(command, args, extras, )
			instance.add_event(event)
		
		return instance

	def to_str(self, *, events_only: bool=False) -> str:
		event_lines = list(map(str, self._events))
		if events_only:
			result = event_lines

		else:
			result = [
				f'timinggroup({self.dump_args_to_str()})' + '{'
			]
			result.extend(f'  {x}' for x in event_lines)
			result.append('};')
		
		return '\n'.join(result)

	def __str__(self) -> str:
		return self.to_str()


	@model_validator(mode='after')
	def _after_validation(self) -> Self:
		if self.noinput is False:
			self.noinput = None

		if self.fadingholds is False:
			self.fadingholds = None

		if self.anglex is not None:
			self.anglex %= 3600
		
		if self.angley is not None:
			self.angley %= 3600

		return self


# :: AUTOGEN-START ::

from .analyse import analyse_annotation, analyse_part_separation, analyse_timinggroup_header, analyse_timinggroup_footer, analyse_command
from .token import TokenType, Token, Tokens, ARG_TOKEN_TYPES, COMMAND_TOKEN_TYPES, TOKEN_TYPES_PATTERNS, tokenize

__all__ = [
    'analyse_annotation', 
    'analyse_part_separation', 
    'analyse_timinggroup_header', 
    'analyse_timinggroup_footer', 
    'analyse_command', 
    'TokenType', 
    'Token', 
    'Tokens', 
    'ARG_TOKEN_TYPES', 
    'COMMAND_TOKEN_TYPES', 
    'TOKEN_TYPES_PATTERNS', 
    'tokenize', 
]

# :: AUTOGEN-END ::



from __future__ import annotations

import json
from io import TextIOWrapper
from typing import Self

from pydantic import Field, NonNegativeInt as uint, PositiveFloat as pfloat, PositiveInt as pint, PrivateAttr, field_serializer, model_validator

from mortis.songlist.base import SonglistPartModel
from mortis.songlist.bgs import Backgrounds
from mortis.songlist.diffs import Difficulties
from mortis.songlist.types import LowerAsciiId, SideEnum, SingleLineStr, StrLocalizedSLRE
from mortis.utils import UnreachableBranch


__all__ = ['SonglistItem']

class SonglistItem(SonglistPartModel):

	# idx: uint
	id: LowerAsciiId
	title_localized: StrLocalizedSLRE

	## mutex: artist & artist_localized
	artist: SingleLineStr | None = None
	artist_localized: StrLocalizedSLRE | None = None

	bpm: str
	bpm_base: pint | pfloat
	pack: LowerAsciiId = Field(alias='set')
	purchase: LowerAsciiId

	side: SideEnum
	@field_serializer('side', when_used='json')
	def _serialize_side(self, v: SideEnum) -> int:
		return v.value
	
	bg: LowerAsciiId
	date: uint
	version: SingleLineStr
	audio_preview: uint = Field(alias='audioPreview')
	audio_preview_end: pint = Field(alias='audioPreviewEnd')

	difficulties: Difficulties


	_relcheck: bool = PrivateAttr(default=True)
	def _enable_relcheck(self) -> None:
		self._relcheck = True
	def _disable_relcheck(self) -> None:
		self._relcheck = False


	@model_validator(mode='after')
	def _after_validation(self) -> Self:
		if not self._relcheck:
			return self
		
		if self.audio_preview >= self.audio_preview_end:
			raise ValueError(f'\'audioPreviewEnd\' must be strictly later than \'audioPreview\'')
		
		if (
			Backgrounds.is_official_bg(self.bg)
			and not Backgrounds.matches(self.side, self.bg)
		):
			raise ValueError(f'\'bg\' ({self.bg}) does not match the \'side\' ({self.side})')

		
		if self.artist is None and self.artist_localized is None:
			raise ValueError(f'At least one of \'artist\' and \'artist_localized\' should be provided')
		if self.artist is not None and self.artist_localized is not None:
			raise ValueError(f'At most one of \'artist\' and \'artist_localized\' should be provided')

		return self
	
	def __enter__(self) -> Self:
		self._disable_relcheck()
		return self
	
	def __exit__(self, exc_type, exc, tb):
		self._enable_relcheck()
		self._after_validation # trigger model after validator
	
	def use_localized_artist(self) -> None:
		if self.artist_localized is not None:
			return
		with self as ctx:
			assert ctx.artist is not None, UnreachableBranch()
			ctx.artist_localized = StrLocalizedSLRE(en=ctx.artist)
			ctx.artist = None
	
	def use_simple_artist(self) -> None:
		if self.artist is not None:
			return
		with self as ctx:
			assert ctx.artist_localized is not None, UnreachableBranch()
			ctx.artist = ctx.artist_localized.en
			ctx.artist_localized = None
	
	
	@classmethod
	def load_from_path(cls, path, /) -> Self:
		with open(path, 'r', encoding='utf-8') as f:
			return cls.load(f)

	@classmethod
	def load(cls, file: TextIOWrapper, /) -> Self:
		lines = file.read()
		return cls.loads(lines)

	@classmethod
	def loads(cls, lines: str, /) -> Self:
		return cls.model_validate_json(lines)

	def dumps(self, /, *, indent: int | None=None) -> str:
		return json.dumps(self.to_dict(), ensure_ascii=False, indent=indent)

	def dump(self, file: TextIOWrapper, /, *, indent: int | None=None) -> None:
		file.write(self.dumps(indent=indent))
	
	def dump_to_path(self, path, /, *, indent: int | None=None) -> None:
		with open(path, 'w', encoding='utf-8') as f:
			self.dump(f, indent=indent)


from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Literal

from mortis.aff.lexer.token import ARG_TOKEN_TYPES, COMMAND_TOKEN_TYPES, Token, Tokens, TokenType


__all__ = [
	'analyse_annotation',
	'analyse_part_separation',
	'analyse_timinggroup_header',
	'analyse_timinggroup_footer',
	'analyse_command'
]

def analyse_annotation(tokens: Tokens) -> tuple[str, Any]:
	if any([
		len(tokens) != 3,
		tokens[0].type_ != TokenType.Identifier,
		tokens[1].type_ != TokenType.Colon,
		tokens[2].type_ not in ARG_TOKEN_TYPES
	]):
		raise ValueError('Invalid annotation format')
	return tokens[0].value, tokens[2].value # type: ignore


def analyse_part_separation(tokens: Tokens) -> None:
	if len(tokens) != 1 or tokens[0].type_ != TokenType.PartSeparator:
		raise ValueError('Invalid part separation format')


def analyse_timinggroup_header(tokens: Tokens) -> str | None:
	exc = ValueError('Invalid timing group header format')
	tokenc = len(tokens)

	if tokenc == 4:
		tg, lparen, rparen, lbrace = tokens
	elif tokenc == 5:
		tg, lparen, params, rparen, lbrace = tokens
	else:
		raise exc
	
	if any([
		tg.type_ != TokenType.TimingGroup,
		lparen.type_ != TokenType.LeftParenthesis,
		(tokenc == 5 and params.type_ != TokenType.Identifier), # type: ignore
		rparen.type_ != TokenType.RightParenthesis,
		lbrace.type_ != TokenType.LeftBrace,
	]):
		raise exc
	
	return params.value if tokenc == 5 else None # type: ignore


def analyse_timinggroup_footer(tokens: Tokens) -> None:
	if any([
		len(tokens) != 2,
		tokens[0].type_ != TokenType.RightBrace,
		tokens[1].type_ != TokenType.Semicolon,
	]):
		raise ValueError('Invalid timing group footer format')


def find_token_of_type(
	tokens: Iterable[Token],
	type_: TokenType,
	start: int = 0
) -> int | None:
	
	tokens = list(tokens)
	for i in range(start, len(tokens)):
		if tokens[i].type_ == type_:
			return i
	return None


def analyse_command(tokens: Tokens) -> tuple[str, tuple, tuple | None]:
	exc = ValueError('Invalid command format')
	if not tokens:
		raise exc
	
	lpidx = find_token_of_type(tokens, TokenType.LeftParenthesis)
	if lpidx is None:
		raise exc
	
	if lpidx == 0:
		command_name = ''
		_, *rest = tokens
	else:
		command, _, *rest = tokens
		if command.type_ not in COMMAND_TOKEN_TYPES:
			raise exc
		command_name = command.value
	
	rpidx = find_token_of_type(rest, TokenType.RightParenthesis)
	if rpidx is None:
		raise exc
	
	argtoks = rest[:rpidx]
	after_rparen_tokens = rest[rpidx + 1:] # type: ignore
	
	if not argtoks:
		raise exc
	
	args = []
	arg_status: Literal['arg', 'comma']	= 'arg'
	while argtoks:
		tok = argtoks.pop(0)
		if arg_status == 'arg' and tok.type_ in ARG_TOKEN_TYPES:
			args.append(tok.value)
			arg_status = 'comma'
		elif arg_status == 'comma' and tok.type_ == TokenType.Comma:
			arg_status = 'arg'
		else:
			raise exc
	
	*rest, trailing = after_rparen_tokens
	if trailing.type_ != TokenType.Semicolon:
		raise exc
	
	if not rest:
		return command_name, tuple(args), None # type: ignore

	restc = len(rest)
	if restc < 3:
		raise exc
	
	lbracket, *extoks, rbracket = rest
	if any([
		lbracket.type_ != TokenType.LeftBracket,
		rbracket.type_ != TokenType.RightBracket,
		len(extoks) % 5 != 4,
	]):
		raise exc
	
	extra_args: list = []
	ex_status = 0
	ex_status_types = {
		0: {TokenType.At, TokenType.Arctap},
		1: {TokenType.LeftParenthesis},
		2: {TokenType.Integer},
		3: {TokenType.RightParenthesis},
		4: {TokenType.Comma},
	}
	while extoks:
		tok = extoks.pop(0)
		if tok.type_ not in ex_status_types[ex_status]:
			raise exc
		
		if ex_status == 2:
			extra_args.append(tok.value)
		
		ex_status = (ex_status + 1) % 5
	
	return command.value, tuple(args), tuple(extra_args) # type: ignore
		


# :: AUTOGEN-START ::

from .bgs import Backgrounds
from .diffs import Difficulty, Difficulties
from .item import SonglistItem
from .types import LowerAsciiId, SingleLineStr, Localized, LocalizedReqEn, StrLocalizedSLRE, SideEnum, RatingClassEnum, RTCLS_STR_MAP, RatingInt
from .utils import get_audio_name, get_preview_name, get_jacket_names, get_aff_name

__all__ = [
    'Backgrounds', 
    'Difficulty', 
    'Difficulties', 
    'SonglistItem', 
    'LowerAsciiId', 
    'SingleLineStr', 
    'Localized', 
    'LocalizedReqEn', 
    'StrLocalizedSLRE', 
    'SideEnum', 
    'RatingClassEnum', 
    'RTCLS_STR_MAP', 
    'RatingInt', 
    'get_audio_name', 
    'get_preview_name', 
    'get_jacket_names', 
    'get_aff_name', 
]

# :: AUTOGEN-END ::



from __future__ import annotations

from enum import StrEnum
from re import compile as compile_regex
from typing import Any


__all__ = [
	'TokenType', 'Token', 'Tokens',
	'ARG_TOKEN_TYPES', 'COMMAND_TOKEN_TYPES', 'TOKEN_TYPES_PATTERNS',
	'tokenize'
]

class TokenType(StrEnum):
	# Grammatical symbols
	LeftParenthesis = r'\('
	RightParenthesis = r'\)'
	LeftBracket = r'\['
	RightBracket = r'\]'
	LeftBrace = r'\{'
	RightBrace = r'\}'
	Comma = r','
	Colon = r':'
	Semicolon = r';'

	# Keywords
	Arctap = r'arctap'
	Arc = r'arc'
	At = r'at'
	Camera = r'camera'
	Hold = r'hold'
	SceneControl = r'scenecontrol'
	TimingGroup = r'timinggroup'
	Timing = r'timing'

	# Types
	Float = r'[+-]?(\d+\.\d*|\.\d+)'
	Integer = r'[+-]?[0-9]+'
	Identifier = r'[a-zA-Z_][a-zA-Z0-9]*'

	# Special
	PartSeparator = r'-'
	Whitespace = r'\s+'

ARG_TOKEN_TYPES = (TokenType.Integer, TokenType.Float, TokenType.Identifier)
COMMAND_TOKEN_TYPES = (
	TokenType.Arc,
	TokenType.Camera,
	TokenType.Hold,
	TokenType.SceneControl,
	TokenType.Timing,
)

TOKEN_TYPES_PATTERNS = [
	(token_type, compile_regex(regex)) 
	for token_type, regex
	in TokenType.__members__.items()
]

class Token:
	def __init__(self, type_: TokenType, value: Any | None = None) -> None:
		self.type_ = type_
		self.value = value
	
	def __repr__(self) -> str:
		value_str = '' if self.value is None else f', {self.value!r}'
		return f'Token({self.type_.name}{value_str})'
	
	def __eq__(self, other) -> bool:
		return type(other) is type(self) and self.type_ == other.type_ and self.value == other.value

Tokens = tuple[Token, ...]

def tokenize(line: str) -> tuple[Token, ...]:
	if not isinstance(line, str):
		raise ValueError(f'Value must be of type str, got value {line!r} of type {type(line)!r}')
	tokens = []
	pos = 0
	line_len = len(line)

	
	while pos < line_len:
		matched = None
		token_type = None
		
		for name, pattern in TOKEN_TYPES_PATTERNS:
			matched = pattern.match(line, pos)
			if matched:
				token_type = TokenType[name]
				break
		else:
			raise ValueError(f"Failed to tokenize at position {pos} (at char {line[pos]!r}): {line!r}")
		
		matched_text = matched.group(0)
		
		if token_type == TokenType.Whitespace:
			pos = matched.end()
			continue
		
		value = None
		if token_type == TokenType.Float:
			value = float(matched_text)

		elif token_type == TokenType.Integer:
			value = int(matched_text)
		
		elif token_type == TokenType.Identifier:
			value = str(matched_text)
		
		else:
			value = matched_text
		
		tokens.append(Token(token_type, value))
		pos = matched.end()
	
	return tuple(tokens)


from __future__ import annotations

from mortis.songlist.diffs import Difficulty


__all__ = [
	'get_audio_name',
	'get_preview_name',
	'get_jacket_names',
	'get_aff_name'
]

def get_audio_name(diff: Difficulty) -> str:
	return f'{diff.rating_class.value}.ogg' if diff.audio_override else 'base.ogg'

def get_preview_name(diff: Difficulty) -> str:
	return f'{diff.rating_class.value}_preview.ogg' if diff.audio_override else 'preview.ogg'

def get_jacket_names(diff: Difficulty) -> tuple[str, ...]:
	rtcls = diff.rating_class if diff.jacket_override else 'base'
	return f'{rtcls}.jpg', f'{rtcls}_256.jpg', f'1080_{rtcls}.jpg', f'1080_{rtcls}_256.jpg'

def get_aff_name(diff: Difficulty) -> str:
	return f'{diff.rating_class.value}.aff'


from __future__ import annotations

from math import isclose
from typing import ClassVar, Self

from pydantic import NonNegativeInt as uint

from mortis.aff.types.models import AFFEventConfig, TechnicalEvent
from mortis.aff.types.dffloat import dffloat2


__all__ = ['Timing']

class Timing(TechnicalEvent):
	
	__aff_config__: ClassVar[AFFEventConfig] = AFFEventConfig(
		is_event=True,
		commands=('timing', ),
		argument_order=('time', 'bpm', 'beat')
	)

	time: uint
	bpm: dffloat2
	beat: dffloat2

	def _after_validation(self) -> Self:
		super()._after_validation()
		
		if self.beat < 0:
			raise ValueError(f'beat must be non-negative')
		
		if not isclose(self.bpm, 0, abs_tol=1e-2) and isclose(self.beat, 0, abs_tol=1e-2):
			raise ValueError(f'beat cannot be zero if bpm is non-zero')
		
		if self.time == 0:
			if self.bpm < 0:
				raise ValueError(f'bpm of initial timing (time == 0) must be non-negative')
			if self.beat < 0:
				raise ValueError(f'beat of initial timing (time == 0) must be non-negative')
		
		return self
	
	@property
	def beat_duration_ms(self) -> float:
		return 6e4 / self.bpm
	
	@property
	def measure_duration_ms(self) -> float:
		return self.beat_duration_ms * self.beat


from __future__ import annotations

from collections.abc import Iterable, Iterator
from io import TextIOWrapper
from typing import Any, Literal, Self

from pydantic import BaseModel, model_validator

from mortis.aff.lexer.analyse import analyse_annotation, analyse_command, analyse_part_separation, analyse_timinggroup_footer, analyse_timinggroup_header
from mortis.aff.lexer.token import tokenize
from mortis.aff.types.models import AFFEvent
from mortis.aff.timinggroup import TimingGroup, parse_event
from mortis.utils import get_default_model_cfg


__all__ = ['AFF']

AudioOffset = 'AudioOffset'
TPDF = 'TimingPointDensityFactor'
class AFF(BaseModel):

	model_config = get_default_model_cfg()

	attrs: dict[str, Any] = {AudioOffset: 0}
	base_group: TimingGroup
	other_groups: list[TimingGroup] = []

	@property
	def offset(self) -> int:
		return self.attrs[AudioOffset]
	
	@offset.setter
	def offset(self, value: int) -> None:
		try:
			self.attrs[AudioOffset]  = int(value)
		except (ValueError, TypeError) as e:
			raise ValueError(e)
	
	@offset.deleter
	def offset(self) -> None:
		self.attrs[AudioOffset] = 0


	def iter_groups(self) -> Iterator[TimingGroup]:
		yield self.base_group
		yield from self.other_groups

	@property
	def groups(self) -> tuple[TimingGroup, ...]:
		return tuple(self.iter_groups())

	def iter_events(self) -> Iterator[AFFEvent]:
		for group in self.iter_groups():
			yield from group.iter_events()
	
	@property
	def events(self) -> tuple[AFFEvent, ...]:
		return tuple(self.iter_events())

	@property
	def tpdf(self) -> float | None:
		return self.attrs.get(TPDF)
	
	@tpdf.setter
	def tpdf(self, value: float) -> None:
		try:
			self.attrs[TPDF] = float(value)
		except (ValueError, TypeError) as e:
			raise ValueError(e)
	
	@tpdf.deleter
	def tpdf(self) -> None:
		if TPDF not in self.attrs:
			return
		del self.attrs[TPDF]

	def unwrap_tpdf(self) -> float:
		return 1.0 if self.tpdf is None else self.tpdf
	
	@classmethod
	def load_from_path(cls, path) -> Self:
		with open(path, 'r', encoding='utf-8') as f:
			return cls.load(f)

	@classmethod
	def load(cls, file: TextIOWrapper) -> Self:
		lines = file.readlines()
		return cls.load_lines(lines)
	
	@classmethod
	def loads(cls, line: str) -> Self:
		return cls.load_lines(line.split('\n'))

	@classmethod
	def load_lines(cls, lines: Iterable[str]) -> Self:
		attrs: dict[str, Any] = {}
		errors: dict[int, Exception] = {}

		base_group: TimingGroup = TimingGroup()
		other_groups: list[TimingGroup] = []
		current_group: TimingGroup = TimingGroup()

		status: Literal['anno', 'normal', 'ingroup'] = 'anno'
		for i, line in enumerate(lines):
			tokens = tokenize(line)

			if status == 'anno':
				try:
					analyse_part_separation(tokens)
					status = 'normal'
					continue
				except ValueError:
					pass

				try:
					anno_name, anno_val = analyse_annotation(tokens)
					attrs[anno_name] = anno_val
				except ValueError as e:
					errors[i] = e

			elif status == 'normal':
				try:
					header_args = analyse_timinggroup_header(tokens)
					current_group = TimingGroup()
					current_group.args = header_args
					status = 'ingroup'
					continue
				except ValueError:
					pass

				try:
					command, args, extras = analyse_command(tokens)
					event = parse_event(command, args, extras)
					base_group.add_event(event)
				except ValueError as e:
					errors[i] = e

			elif status == 'ingroup':
				try:
					analyse_timinggroup_footer(tokens)
					other_groups.append(current_group)
					status = 'normal'
					continue
				except ValueError:
					pass

				try:
					command, args, extras = analyse_command(tokens)
					event = parse_event(command, args, extras)
					current_group.add_event(event)
				except ValueError as e:
					errors[i] = e
		
		if errors:
			raise ValueError(f'Errors occurred during parsing. Details are: \n{errors!r}')
		
		return cls(
			attrs = attrs,
			base_group = base_group,
			other_groups = other_groups,
		)
	
	def dump_lines(self) -> list[str]:
		lines = [f'{attr}:{value}' for attr, value in self.attrs.items()]
		lines.append('-')
		lines.append(self.base_group.to_str(events_only=True))
		lines.extend(list(map(str, self.other_groups)))
		return lines
	
	def dumps(self) -> str:
		return '\n'.join(self.dump_lines())

	def dump(self, file: TextIOWrapper) -> None:
		file.write(self.dumps())
	
	def dump_to_path(self, path) -> None:
		with open(path, 'w', encoding='utf-8') as f:
			self.dump(f)
	
	def __str__(self) -> str:
		return self.dumps()
	

	@property
	def required_hitsounds(self) -> set[str]:
		hitsounds = set()
		for group in self.groups:
			hitsounds |= group.required_hitsounds
		return hitsounds

	
	@model_validator(mode='after')
	def _after_validation(self) -> Self:
		self.offset = self.attrs.get(AudioOffset, 0)
		
		tpdf = self.attrs.get('TimingPointDensityFactor')
		if tpdf is None:
			del self.tpdf
		else:
			self.tpdf = tpdf
		
		return self
	


from __future__ import annotations

from mortis.songlist.types import SideEnum
from mortis.utils import classproperty


__all__ = ['Backgrounds']

class Backgrounds:
	data = {
		SideEnum.Light: (),
		SideEnum.Conflict: (),
		SideEnum.Achromatic: (),
		SideEnum.Lephon: (),
	}

	@classmethod
	def is_official_bg(cls, bg: str) -> bool:
		return any(bg in bgs for bgs in cls.data.values())

	@classmethod
	def matches(cls, side: SideEnum, bg: str) -> bool:
		return bg in cls.data.get(side, ())
	
	@classproperty
	@classmethod
	def light_bgs(cls) -> tuple[str, ...]:
		return cls.data[SideEnum.Light]
	
	@classproperty
	@classmethod
	def conflict_bgs(cls) -> tuple[str, ...]:
		return cls.data[SideEnum.Conflict]
	
	@classproperty
	@classmethod
	def achromatic_bgs(cls) -> tuple[str, ...]:
		return cls.data[SideEnum.Achromatic]
	
	@classproperty
	@classmethod
	def lephon_bgs(cls) -> tuple[str, ...]:
		return cls.data[SideEnum.Lephon]


from __future__ import annotations

from typing import ClassVar

from pydantic import NonNegativeInt as uint

from mortis.aff.types.models import AFFEventConfig, FloorEvent, TapLikeEvent
from mortis.aff.types.coord import Coordinate, FixedLane, FloatLane


__all__ = ['Tap', 'TapFloat']

class Tap(FloorEvent, TapLikeEvent):
	__aff_config__: ClassVar[AFFEventConfig] = AFFEventConfig(
		is_event=True,
		commands=('', ),
		argument_order=('time', 'lane')
	)

	time: uint
	lane: FixedLane

	def to_fixed_lane(self) -> Tap:
		return self

	def to_float_lane(self) -> TapFloat:
		return TapFloat(
			time = self.time,
			lane = Coordinate.fixed_to_float(self.lane),
		)
	
	def mirror(self) -> None:
		self.lane = Coordinate.mirror_fixed(self.lane)


class TapFloat(Tap):

	lane: FloatLane

	def to_float_lane(self) -> TapFloat:
		return self

	def to_fixed_lane(self) -> Tap:
		return Tap(
			time = self.time,
			lane = Coordinate.float_to_fixed(self.lane)
		)
	
	def mirror(self) -> None:
		self.lane = Coordinate.mirror_float(self.lane) # type: ignore


from __future__ import annotations

from enum import IntEnum
from typing import Annotated, Any, Generic, TypeVar, final
from typing_extensions import Self

import annotated_types
from pydantic import AfterValidator, Field, model_validator

from mortis.globcfg import GlobalConfig
from mortis.songlist.base import SonglistPartModel


__all__ = [
	'LowerAsciiId', 'SingleLineStr',
	'Localized', 'LocalizedReqEn', 'StrLocalizedSLRE',
	'SideEnum', 'RatingClassEnum',
	'RTCLS_STR_MAP', 'RatingInt'
]


def ensure_lower_ascii_id(s: str) -> str:
	if s.isidentifier() and s.isascii() and s.lower() == s:
		return s
	raise ValueError(f'Value must be a lowercase ASCII identifier')
LowerAsciiId = Annotated[str, AfterValidator(ensure_lower_ascii_id)]


def ensure_single_line(s: str) -> str:
	if '\n' not in s:
		return s
	raise ValueError(f'Value must be in one line (contain no \'\\n\')')
SingleLineStr = Annotated[str, AfterValidator(ensure_single_line)]


@final
class GuardinaError(ValueError, KeyError):
	__instance__ = None
	def __new__(cls) -> Self:
		if cls.__instance__ is None:
			cls.__instance__ = super().__new__(cls)
		return cls.__instance__
	
	def __init__(self) -> None:
		pass

	def __str__(self) -> str:
		return f'\'kr\' is not a valid language code; do you mean \'ko\'?'
	
	def __repr__(self) -> str:
		return f'{self.__class__.__name__}()'

T = TypeVar('T')
class Localized(SonglistPartModel, Generic[T]):
	en: T | None = None
	ja: T | None = None
	ko: T | None = None
	zh_Hans: T | None = Field(default=None, alias='zh-Hans')
	zh_Hant: T | None = Field(default=None, alias='zh-Hant')

	@model_validator(mode='before')
	def _before_validation(cls, data: Any) -> Any:
		if not isinstance(data, dict):
			return data
		if 'kr' in data:
			if GlobalConfig.allows_kr_langcode:
				if 'ko' not in data:
					data['ko'] = data['kr']
				del data['kr']
			else:
				raise GuardinaError
		return data

	def __bool__(self) -> bool:
		return bool(self.nondefault_fields)

class LocalizedReqEn(Localized[T]):
	en: T  # pyright: ignore[reportGeneralTypeIssues, reportIncompatibleVariableOverride] | intended behaviour

	def __bool__(self) -> bool:
		return True

StrLocalizedSLRE = LocalizedReqEn[SingleLineStr]


class SideEnum(IntEnum):
	Light = 0
	Conflict = 1
	Achromatic = 2
	Lephon = 3

class RatingClassEnum(IntEnum):
	Past = 0
	Present = 1
	Future = 2
	Beyond = 3
	Eternal = 4
RTCLS_STR_MAP = {e: e.name.lower() for e in RatingClassEnum}


RatingInt = Annotated[int, annotated_types.Ge(-1)]



# :: AUTOGEN-START ::

from .coord import FixedLane, FloatLane, ArcCoord, Coordinate
from .dffloat import dffloat, dffloat2, dffloat3
from .easings import BaseEasing, EasingLinear, EasingSineIn, EasingSineOut, EasingBezierDefault, EasingSineInOut, ArcEasing, get_easing_x, get_easing_y
from .models import AFFEventConfig, AFFEvent, GameObjectEvent, TechnicalEvent, FloorEvent, SkyEvent, LongNoteEvent, TapLikeEvent

__all__ = [
    'FixedLane', 
    'FloatLane', 
    'ArcCoord', 
    'Coordinate', 
    'dffloat', 
    'dffloat2', 
    'dffloat3', 
    'BaseEasing', 
    'EasingLinear', 
    'EasingSineIn', 
    'EasingSineOut', 
    'EasingBezierDefault', 
    'EasingSineInOut', 
    'ArcEasing', 
    'get_easing_x', 
    'get_easing_y', 
    'AFFEventConfig', 
    'AFFEvent', 
    'GameObjectEvent', 
    'TechnicalEvent', 
    'FloorEvent', 
    'SkyEvent', 
    'LongNoteEvent', 
    'TapLikeEvent', 
]

# :: AUTOGEN-END ::



from __future__ import annotations

from enum import Enum
from typing import Any

from pydantic import BaseModel, ConfigDict

__all__ = []

class SonglistPartModel(BaseModel):
	model_config = ConfigDict(
		extra='forbid',
		allow_inf_nan=False,
		validate_by_alias=True,
		serialize_by_alias=True,
		validate_default=True,
		validate_assignment=True,
		json_encoders={Enum: lambda e: e.value}
	)

	def to_dict(self) -> dict:
		return self.model_dump(exclude_defaults=True, by_alias=True, mode='json')
	
	@property
	def nondefault_fields(self) -> dict[str, Any]:
		result = {}
		for name, info in self.__class__.model_fields.items():
			attr = getattr(self, name)
			if info.is_required() or attr != info.default:
				result[name] = attr
		return result


from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, Literal, Self, TypeVar, final

from pydantic import ConfigDict


__all__ = ['get_default_model_cfg']

MT = TypeVar('MT', bound=Any)
RT = TypeVar('RT')
class classproperty(Generic[MT, RT]):
	"""
	Class property attribute. 
	"""

	def __init__(self, fget: Callable[[type[MT]], RT], /) -> None:
		self.fget = classmethod(fget)

	def __get__(self, instance: MT | None, objtype: type[MT], /) -> RT:
		return self.fget.__get__(instance, objtype)()


def get_default_model_cfg(
	extra: Literal['allow', 'ignore', 'forbid'] = 'forbid',
	ignored_types: Any = None
) -> ConfigDict:
	
	if ignored_types is None:
		ignored_types = ()

	return ConfigDict(
		extra=extra,
		validate_assignment=True,
		validate_default=True,
		ignored_types=ignored_types + (classproperty, ),
		allow_inf_nan=False,
	)


@final
class UnreachableBranch(Exception):
	__instance__ = None
	
	def __new__(cls) -> Self:
		if cls.__instance__ is None:
			cls.__instance__ = super().__new__(cls)
		return cls.__instance__
	
	def __init__(self) -> None:
		pass

	def __str__(self) -> str:
		return 'Current code branch is expected to be unreachable'
	
	def __repr__(self) -> str:
		return '<UnreachableBranch>'
UnreachableBranch() # create singleton


from __future__ import annotations

from enum import StrEnum
from typing import ClassVar

from pydantic import NonNegativeInt as uint

from mortis.aff.types.models import AFFEventConfig, TechnicalEvent
from mortis.aff.types.dffloat import dffloat2


__all__ = ['CameraEasing', 'Camera']

class CameraEasing(StrEnum):

	Linear = s = 's'
	CubicIn = qi = 'qi'
	CubicOut = qo = 'qo'
	Reset = reset = 'reset'


class Camera(TechnicalEvent):

	__aff_config__: ClassVar[AFFEventConfig] = AFFEventConfig(
		is_event=True,
		commands=('camera', ),
		argument_order=(
			'time',
			'truck', 'pedestral', 'dolly',
			'pan', 'tilt', 'roll',
			'easing', 'duration'
		)
	)
	
	time: uint
	truck: dffloat2
	"""Movement along the X-axis"""
	pedestral: dffloat2
	"""Movement along the Y-axis"""
	dolly: dffloat2
	"""Movement along the Z-axis"""
	pan: dffloat2
	"""Rotation around the Y-axis"""
	tilt: dffloat2
	"""Rotation around the X-axis"""
	roll: dffloat2
	"""Rotation around the Z-axis"""
	easing: CameraEasing
	duration: uint


from __future__ import annotations

from typing import Any, ClassVar, Self, final

from pydantic import BaseModel, ValidationError, model_validator
from pydantic.fields import FieldInfo
from pydantic_core.core_schema import CoreSchema, union_schema, no_info_plain_validator_function

from mortis.aff.lexer.analyse import analyse_command
from mortis.aff.lexer.token import tokenize
from mortis.utils import get_default_model_cfg, classproperty

__all__ = [
	'AFFEventConfig',
	'AFFEvent',
	'GameObjectEvent', 'TechnicalEvent',
	'FloorEvent', 'SkyEvent', 'LongNoteEvent', 'TapLikeEvent'
]

class AFFEventConfig:
	def __init__(
		self,
		*,
		is_event: bool = False,
		commands: tuple[str, ...] | None = None,
		trailing: str = ';',
		uses_extra_args: bool = False,
		argument_order: tuple[str, ...] | None = None
	) -> None:
		
		self.is_event = is_event
		self.commands = commands if commands is not None else ()
		self.trailing = trailing
		self.uses_extra = uses_extra_args
		self.fields_order = argument_order if argument_order is not None else ()

class AFFEvent(BaseModel):

	model_config = get_default_model_cfg(ignored_types=(AFFEventConfig, ))

	__aff_config__: ClassVar[AFFEventConfig] = AFFEventConfig()
	__converters__: ClassVar[tuple[str, ...]] = 'from_str',
	
	########################################################################################

	@classmethod
	def is_event_class(cls) -> bool:
		return cls.__aff_config__.is_event
	
	@classproperty
	def commands(cls) -> tuple[str, ...]:
		if not cls.is_event_class():
			raise TypeError(f'{cls.__name__} is not an event class')
		
		return cls.__aff_config__.commands
	
	@classproperty
	def main_command(cls) -> str:
		if not cls.is_event_class():
			raise TypeError(f'{cls.__name__} is not an event class')
		
		assert cls.__aff_config__.commands, f'{cls.__name__} should have at least one command name'
		return cls.__aff_config__.commands[0]
	
	@classproperty
	def command_trailing(cls) -> str:
		if not cls.is_event_class():
			raise TypeError(f'{cls.__name__} is not an event class')
		
		return cls.__aff_config__.trailing
	
	@classmethod
	def allows_extra(cls) -> bool:
		if not cls.is_event_class():
			raise TypeError(f'{cls.__name__} is not an event class')
		
		return cls.__aff_config__.uses_extra
	
	@classproperty
	def fields_order(cls) -> tuple[str, ...]:
		if not cls.is_event_class():
			raise TypeError(f'{cls.__name__} is not an event class')
		
		return cls.__aff_config__.fields_order
	
	########################################################################################

	@classproperty
	def model_ordered_fields(cls) -> dict[str, FieldInfo]:
		return {
			name: cls.model_fields[name]
			for name in cls.fields_order
		}

	@classproperty
	def model_required_fields(cls) -> dict[str, FieldInfo]:
		return {
			name: info
			for name, info in cls.model_ordered_fields.items()
			if info.is_required()
		}

	@classproperty
	def model_optional_fields(cls) -> dict[str, FieldInfo]:
		return {
			name: info
			for name, info in cls.model_ordered_fields.items()
			if not info.is_required()
		}
	
	@property
	def ordered_fields(self) -> dict[str, Any]:
		return {
			name: getattr(self, name)
			for name in self.__class__.model_ordered_fields
		}
	
	@property
	def required_fields(self) -> dict[str, Any]:
		return {
			name: getattr(self, name)
			for name in self.__class__.model_required_fields
		}
	
	@property
	def optional_fields(self) -> dict[str, Any]:
		return {
			name: getattr(self, name)
			for name in self.__class__.model_optional_fields
		}

	@property
	def nondefault_fields(self) -> dict[str, Any]:
		result = self.required_fields
		result.update({
			name: current
			for name, info in self.__class__.model_ordered_fields.items()
			if (current := getattr(self, name)) != info.default
		})
		return result
	
	@classproperty
	def min_fields_count(cls) -> int:
		return len(cls.model_required_fields)
	
	@classproperty
	def max_fields_count(cls) -> int:
		return len(cls.model_ordered_fields)
	
	########################################################################################
	
	def __eq__(self, another: Any) -> bool:
		return type(another) is type(self) and self.to_dict() == another.to_dict()
	
	def __ne__(self, another: Any) -> bool:
		return not (self == another)
	
	########################################################################################

	def copy(self) -> Self: # type: ignore
		return self.model_copy()
	
	def deepcopy(self) -> Self:
		return self.model_copy(deep=True)
	
	########################################################################################

	def _set_extras(self, extras: Any) -> None:
		pass

	def _get_extras(self) -> Any:
		return None
	
	########################################################################################

	@final
	@model_validator(mode='before')
	@classmethod
	def __model_validator_before__(cls, data: Any) -> Any:
		return cls._before_validation(data)
	
	@final
	@model_validator(mode='after')
	def __model_validator_after__(self) -> Self:
		return self._after_validation()
	
	@classmethod
	def _before_validation(cls, data: Any) -> Any:
		return data
	
	def _after_validation(self) -> Self:
		return self
	
	########################################################################################

	@classmethod
	def from_dict(cls, data: dict) -> Self:
		extras = {}
		if 'extras' in data:
			extras = data.pop('extras')

		instance = cls.model_validate(data)
		if extras and cls.allows_extra():
			instance._set_extras(extras)

		return instance
	
	def to_dict(self) -> dict:
		data = self.nondefault_fields.copy()
		if self.allows_extra():
			extras = self._get_extras()
			if extras:
				data['extras'] = extras
		return data

	@classmethod
	def from_str(cls, line: str) -> Self:
		if not cls.is_event_class():
			raise ValueError(f'{cls.__name__} is not a event class')
		
		tokens = tokenize(line)
		command, args, extras = analyse_command(tokens)
		return cls.from_analysed(command, args, extras)

	def to_str(self) -> str:
		cls = self.__class__
		if not cls.is_event_class():
			raise TypeError(f'{cls.__name__} is not a event class')
		
		main_command = cls.main_command
		args = []
		for arg in cls.fields_order:
			info = cls.model_ordered_fields[arg]
			current = getattr(self, arg)
			if not info.is_required() and current == info.default:
				break
			args.append(current)
		
		args_str = '(' + ','.join(list(map(str, args))) + ')'

		extras_str = ''		
		if cls.allows_extra():
			extras = self._get_extras()
			if extras:
				extras_str = '[' + ','.join(list(map(str, extras))) + ']'

		
		return main_command + args_str + extras_str + cls.command_trailing
	
	def __str__(self) -> str:
		return self.to_str()
	
	########################################################################################

	@classmethod
	def from_analysed(cls, command: str, args: tuple, extras: tuple | None = None) -> Self:
		if not cls.is_event_class():
			raise ValueError(f'{cls.__name__} is not a event class')
		
		if command not in cls.commands:
			raise ValueError(f'Command name does not match; expected any of {cls.commands!r}, got {command}')
		
		return cls.from_args(args, extras)
	
	@classmethod
	def from_args(cls, args: tuple, extras: tuple | None = None) -> Self:
		if not cls.is_event_class():
			raise ValueError(f'{cls.__name__} is not a event class')
		
		argc = len(args)
		maxc = cls.max_fields_count
		minc = cls.min_fields_count
		if argc > maxc or argc < minc:
			argc_str = (
				str(maxc) if maxc == minc
				else f'{minc} to {maxc}'
			)
			raise ValueError(f'invalid format; expected {argc_str} arguments, got {argc}')
		
		data: dict = {name: arg for name, arg in zip(cls.fields_order, args)}
		instance = cls.model_validate(data)

		if extras:
			instance._set_extras(extras)

		return instance
	
	########################################################################################

	@final
	@classmethod
	def from_any(cls, data: Any) -> Self:
		errors: list = []

		for name in cls.__converters__:
			func = getattr(cls, name)
			try:
				return func(data)
		
			except ValidationError as e:
				for error in e.errors():
					if 'ctx' not in error:
						error['ctx'] = {}
					
					error['ctx']['converter'] = name
					error['ctx']['error'] = e
					errors.append(error) # type: ignore

			except (ValueError, TypeError) as e:
				errors.append({
					'type': 'value_error',
					'input': data,
					'ctx': {'converter': name, 'error': e}
				})
		
		if errors:
			raise ValidationError.from_exception_data(
				title = f"instantiating {cls.__name__} from {data}",
				line_errors = errors
			)
		
		assert False, 'never reaches here'
	
	########################################################################################
		
	@final
	@classmethod
	def __get_pydantic_core_schema__(cls, source_type, handler) -> CoreSchema:
		converters = [handler(source_type)]
		converters.extend(
			no_info_plain_validator_function(getattr(cls, name))
			for name in cls.__converters__
		)
		return union_schema(converters)

class GameObjectEvent(AFFEvent):
	def mirror(self) -> None:
		pass

class TechnicalEvent(AFFEvent):
	pass

class FloorEvent(GameObjectEvent):
	pass

class SkyEvent(GameObjectEvent):
	pass

class LongNoteEvent(GameObjectEvent):
	@property
	def time_range(self) -> tuple[int, int]:
		return (self.begin_time, self.end_time)
	
	@time_range.setter
	def time_range(self, value: tuple[int, int]) -> None:
		begin, end = value
		if begin > end:
			raise ValueError(f'Invalid time range; begin should not be greater than end')
		
		if begin <= self.end_time:
			self.begin_time, self.end_time = begin, end
		else:
			self.end_time, self.begin_time = end, begin

	@property
	def duration(self) -> int:
		return self.end_time - self.begin_time
	
	def __len__(self) -> int:
		return self.duration
	
	def _after_validation(self) -> Self:
		super()._after_validation()
		if self.end_time < self.begin_time:
			raise ValueError('end_time must be greater than or equal to begin_time')
		return self

class TapLikeEvent(GameObjectEvent):
	pass


from __future__ import annotations

from enum import StrEnum
from typing import ClassVar, Self

from pydantic import NonNegativeInt as uint

from mortis.aff.types.models import AFFEventConfig, TechnicalEvent
from mortis.aff.types.dffloat import dffloat2


__all__ = ['SceneControlType', 'SceneControl']

class SceneControlType(StrEnum):

	TrackHide = trackhide = 'trackhide'
	TrackShow = trackshow = 'trackshow'
	
	TrackDisplay = trackdisplay = 'trackdisplay'
	RedLine = redline = 'redline'
	ArcahvDistort = arcahvdistort = 'arcahvdistort'
	ArcahvDebris = arcahvdebris = 'arcahvdebris'
	HideGroup = hidegroup = 'hidegroup'
	EnwidenCamera = enwidencamera = 'enwidencamera'
	EnwidenLanes = enwidenlanes = 'enwidenlanes'


class SceneControl(TechnicalEvent):

	__aff_config__: ClassVar[AFFEventConfig] = AFFEventConfig(
		is_event=True,
		commands=('scenecontrol', ),
		argument_order=('time', 'type_', 'duration', 'target_state')
	)
	
	time: uint
	type_: SceneControlType
	duration: dffloat2 | None = None
	target_state: uint | None = None

	def _after_validation(self) -> Self:
		super()._after_validation()

		# scenecontrol(t, type);
		if self.type_ in [SceneControlType.TrackHide, SceneControlType.TrackShow]:
			if self.duration is not None or self.target_state is not None:
				raise ValueError(f'{self.type_} does not accept any extra parameter')
			
			return self
			
		# scenecontrol(t, type, duration, target_state);
		if self.type_ == SceneControlType.HideGroup:
			# unused, but need a non-None value
			if self.duration is None:
				self.duration = dffloat2(1.00)
		else:
			if self.duration is None:
				raise ValueError(f'duration of {self.type_} is necessary')
		
		if self.duration < 0:
			raise ValueError(f'duration must be non-negative, got {self.duration}')
		
		if self.type_ == SceneControlType.RedLine:
			# unused, but need a non-None value.
			if self.target_state is None:
				self.target_state = 0
		else:
			if self.target_state is None:
				raise ValueError(f'target_state of {self.type_} is necessary')
			
			if self.type_ in {
				SceneControlType.EnwidenCamera,
				SceneControlType.EnwidenLanes,
				SceneControlType.HideGroup,
			} and self.target_state not in {0, 1}:
				raise ValueError(
					f'target_state of {self.type_} is a switch value that only accepts 0 or 1, got {self.target_state}'
				)
		
		return self


from __future__ import annotations

from enum import StrEnum
from math import cos, pi, sin
from typing import final


__all__ = [
	'BaseEasing', 'EasingLinear', 'EasingSineIn', 'EasingSineOut', 'EasingBezierDefault', 'EasingSineInOut',
	'ArcEasing', 'get_easing_x', 'get_easing_y'
]

class BaseEasing:
	@staticmethod
	def _at_unit(ratio: float) -> float:
		"""Implementation of the easing curve, typically the y=f(x) expression of it."""
		raise NotImplementedError

	@final
	@classmethod
	def at_unit(cls, ratio: float, /) -> float:
		if ratio < 0 or ratio > 1:
			raise ValueError(f'ratio must be within [0, 1]')
		return cls._at_unit(ratio)
	
	@final
	def at(self, value: float, /) -> float:
		ratio = (value - self.begin_input) / (self.end_input - self.begin_input)
		if ratio < 0 or ratio > 1:
			raise ValueError(f'Param must be within the range between {self.begin_input} and {self.end_input}')
		
		eased_ratio = self.at_unit(ratio)
		return self.begin_input + eased_ratio * (self.end_output - self.begin_output)
	
	def __init__(
			self,
			begin_input: float,
			end_input: float,
			begin_output: float,
			end_output: float,
		):
		self.begin_input = begin_input
		self.end_input = end_input
		self.begin_output = begin_output
		self.end_output = end_output
	
class EasingLinear(BaseEasing):
	@staticmethod
	def _at_unit(ratio: float) -> float:
		return ratio

class EasingSineIn(BaseEasing):
	@staticmethod
	def _at_unit(ratio: float) -> float:
		return sin(pi * ratio / 2)

class EasingSineOut(BaseEasing):
	@staticmethod
	def _at_unit(ratio: float) -> float:
		return 1 - cos(pi * ratio / 2)

class EasingBezierDefault(BaseEasing):
	"""Bezier curve, starts from (0, 0), ends at (1, 1), with 1st and 2nd control points (1/3, 0) and (2/3, 1)."""
	@staticmethod
	def _at_unit(ratio: float) -> float:
		return 3 * ratio ** 2 - 2 * ratio ** 3 # Simplified formula. Gives exact value.

class EasingSineInOut(BaseEasing):
	@staticmethod
	def _at_unit(ratio: float) -> float:
		return (1 - cos(pi * ratio)) / 2
	
class ArcEasing(StrEnum):
	S = s = Straight = 's'
	B = b = Bezier = 'b'
	Si = si = SineIn = 'si'
	So = so = SineOut = 'so'
	SiSi = sisi = 'sisi'
	SiSo = siso = 'siso'
	SoSi = sosi = 'sosi'
	SoSo = soso = 'soso'

def get_easing_x(arc_easing: ArcEasing) -> type[BaseEasing]:
	if arc_easing in [ArcEasing.S]:
		return EasingLinear
	elif arc_easing in [ArcEasing.B]:
		return EasingBezierDefault
	elif arc_easing in [ArcEasing.Si, ArcEasing.SiSi, ArcEasing.SiSo]:
		return EasingSineIn
	elif arc_easing in [ArcEasing.So, ArcEasing.SoSi, ArcEasing.SoSo]:
		return EasingSineOut
	assert False

def get_easing_y(arc_easing: ArcEasing) -> type[BaseEasing]:
	if arc_easing in [ArcEasing.S, ArcEasing.Si, ArcEasing.So]:
		return EasingLinear
	elif arc_easing in [ArcEasing.B]:
		return EasingBezierDefault
	elif arc_easing in [ArcEasing.SiSi, ArcEasing.SoSi]:
		return EasingSineIn
	elif arc_easing in [ArcEasing.SiSo, ArcEasing.SoSo]:
		return EasingSineOut
	assert False


from __future__ import annotations

from collections.abc import Iterable, Iterator
from typing import Any, Self

from pydantic import Field, NonNegativeInt as uint, model_serializer, model_validator

from mortis.songlist.base import SonglistPartModel
from mortis.songlist.types import RTCLS_STR_MAP, LowerAsciiId, RatingClassEnum, RatingInt, SingleLineStr, StrLocalizedSLRE


__all__ = ['Difficulty', 'Difficulties']

class Difficulty(SonglistPartModel):
	rating_class: RatingClassEnum = Field(alias='ratingClass')
	chart_designer: str = Field(alias='chartDesigner')
	jacket_designer: str = Field(alias='jacketDesigner')
	rating: RatingInt
	rating_plus: bool = Field(default=False, alias='ratingPlus')

	title_localized: StrLocalizedSLRE | None = None
	audio_override: bool = Field(default=False, alias='audioOverride')
	jacket_override: bool = Field(default=False, alias='jacketOverride')
	bg: LowerAsciiId | None = None
	date: uint | None = None
	version: SingleLineStr | None = None

	def is_activated(self) -> bool:
		return self.rating != -1
	
	def __bool__(self) -> bool:
		return self.is_activated()

	@property
	def rating_str(self) -> str:
		return f'{self.rating}+' if self.rating_plus else str(self.rating)

	@model_validator(mode='after')
	def _after_validation(self) -> Self:
		if self.rating in (-1, 0) and self.rating_plus:
			raise ValueError(f'{self.rating_str} is not a valid rating level')
		return self

class _DifficultyList(SonglistPartModel):
	data: list[Difficulty]

	@model_validator(mode='before')
	@classmethod
	def _before_validation(cls, value: Any) -> Any:
		if isinstance(value, cls):
			return value
		elif isinstance(value, Iterable) and not isinstance(value, dict):
			return {'data': value}
		return value
	
	@model_validator(mode='after')
	def _after_validation(self) -> Self:
		found = set()
		duplicated: list[RatingClassEnum] = []

		for diff in self.data:
			rtcls = diff.rating_class
			if rtcls in found:
				duplicated.append(rtcls)
			else:
				found.add(rtcls)
		
		if duplicated:
			duplicated_str = ', '.join(e.name for e in duplicated)
			raise ValueError(f'Duplicated difficulties: {duplicated_str}')
		
		return self

	def to_dict(self) -> dict[str, Difficulty]:
		return {RTCLS_STR_MAP[item.rating_class]: item for item in self.data}


class Difficulties(SonglistPartModel):
	past: Difficulty
	present: Difficulty
	future: Difficulty
	beyond: Difficulty | None = None
	eternal: Difficulty | None = None

	@model_validator(mode='before')
	@classmethod
	def _before_validation(cls, value: Any) -> Any:
		if isinstance(value, cls):
			return value
		elif isinstance(value, Iterable) and not isinstance(value, dict):
			diff_list = _DifficultyList.model_validate(value)
			return diff_list.to_dict()
		elif isinstance(value, _DifficultyList):
			return value.to_dict()
		return value
	
	@property
	def all_declared(self) -> tuple[Difficulty, ...]:
		return tuple(self.nondefault_fields.values())

	@property
	def all_activated(self) -> tuple[Difficulty, ...]:
		return tuple(
			diff for name in self.__class__.model_fields
			if (diff := getattr(self, name)) is not None and diff.is_activate()
		)

	@model_serializer
	def serialize(self) -> list[Difficulty]:
		return list(self.all_declared)

	def __getitem__(self, index: RatingClassEnum) -> Difficulty | None:
		index = RatingClassEnum(index)
		name = RTCLS_STR_MAP[index]
		diff = getattr(self, name)
		if diff is None:
			raise IndexError(f'No such difficulty: {index}')
		return diff
	
	def __setitem__(self, index: RatingClassEnum, value: Difficulty) -> None:
		index = RatingClassEnum(index)
		name = RTCLS_STR_MAP[index]
		value = Difficulty.model_validate(value)
		setattr(self, name, value)
	
	def __delitem__(self, index: RatingClassEnum) -> None:
		index = RatingClassEnum(index)
		if index in {RatingClassEnum.Past, RatingClassEnum.Present, RatingClassEnum.Future}:
			raise IndexError(f'Failed to delete difficulty {index.name}: required and undeletable')
		name = RTCLS_STR_MAP[index]
		setattr(self, name, None)
	
	def iter_difficulty(self) -> Iterator[Difficulty]:
		yield from self.all_declared

	def __contains__(self, rtcls: RatingClassEnum) -> bool:
		try:
			self[rtcls]
			return True
		except IndexError:
			return False


from __future__ import annotations

from collections.abc import Iterable, Iterator
from enum import IntEnum, StrEnum
from typing import Any, ClassVar, Self

from pydantic import NonNegativeInt as uint, PrivateAttr

from mortis.aff.types.models import AFFEventConfig, LongNoteEvent, SkyEvent, TapLikeEvent
from mortis.aff.types.coord import ArcCoord, Coordinate
from mortis.aff.types.easings import ArcEasing, BaseEasing, get_easing_x, get_easing_y
from mortis.aff.types.dffloat import dffloat2


__all__ = ['ArcType', 'ArcColor', 'Arctap', 'Arc', 'ScaledArctap']

class ArcType(StrEnum):

	Solid = Snake = False_ = 'false'
	Void = Trace = True_ = 'true'
	Designant = 'designant'


class ArcColor(IntEnum):

	Blue = 0
	Red = 1
	Green = 2
	Gray = Grey = 3


class Arctap(SkyEvent, TapLikeEvent):

	__aff_config__: ClassVar[AFFEventConfig] = AFFEventConfig(
		is_event=True,
		trailing='',
		commands=('arctap', 'at'),
		argument_order=('time',)
	)
	__converters__: ClassVar[tuple[str, ...]] = 'from_int', 'from_str'

	time: uint

	_parent: Arc | None = PrivateAttr(default=None)

	@property
	def parent(self) -> Arc | None:
		return self._parent
	
	@parent.setter
	def parent(self, value: Arc | None) -> None:
		if value is None:
			del self.parent
		else:
			self.bind_to(value)
	
	@parent.deleter
	def parent(self) -> None:
		self.unbind()	

	def bind_to(self, arc: Arc) -> None:
		if not isinstance(arc, Arc):
			raise TypeError('arc must be an instance of Arc')
		
		if self._parent is arc:
			return

		if self._parent is not None:
			self.unbind()

		if self.time > arc.end_time or self.time < arc.begin_time:
			raise ValueError(f'Arctap time {self.time} is not within the range of arc: {arc.time_range}')

		arc._arctaps.append(self)
		arc._arctaps.sort(key=lambda at: at.time)
		self._parent = arc

	def unbind(self) -> None:
		if self._parent is None:
			return
		
		idx = self._parent.find_arctap(self)
		del self._parent._arctaps[idx]
		self._parent = None

	def position(self) -> tuple[float, float]:
		if self._parent is None:
			raise ValueError('Arctap is not bound to any arc')
		return self._parent.position_at(self.time)
	
	@classmethod
	def from_int(cls, value: int) -> Self:
		try:
			return cls(time=int(value))
		except (ValueError, TypeError) as e:
			raise ValueError(e)
	
	def copy(self) -> Self:
		return self.__copy__()
	
	def deepcopy(self) -> Self:
		return self.__deepcopy__()
	
	def __copy__(self) -> Self:
		return self.__class__(time=self.time)
	
	def __deepcopy__(self, memo: dict[int, Any] | None = None) -> Self:
		instance = self.__class__(time=self.time)
		instance.parent = self.parent
		return instance


class Arc(SkyEvent, LongNoteEvent):

	__aff_config__: ClassVar[AFFEventConfig] = AFFEventConfig(
		is_event=True,
		uses_extra_args=True,
		commands=('arc', ),
		argument_order=(
			'begin_time', 'end_time',
			'begin_x', 'end_x',
			'easing',
			'begin_y', 'end_y',
			'color', 'hitsound', 'type_',
			'smoothness'
		)
	)
	
	begin_time: uint
	end_time: uint
	begin_x: ArcCoord
	end_x: ArcCoord
	easing: ArcEasing
	begin_y: ArcCoord
	end_y: ArcCoord
	color: ArcColor
	hitsound: str
	type_: ArcType
	smoothness: dffloat2 | None = None

	_arctaps: list[Arctap] = []
	
	@property
	def easing_x(self) -> BaseEasing:
		return get_easing_x(self.easing)(self.begin_time, self.end_time, self.begin_x, self.end_x)
	
	@property
	def easing_y(self) -> BaseEasing:
		return get_easing_y(self.easing)(self.begin_time, self.end_time, self.begin_y, self.end_y)
	

	def wrap_hitsound(self, hitsound: str | None) -> None:
		self.hitsound = 'none' if hitsound is None else hitsound.replace('.', '_')
	
	def unwrap_hitsound(self) -> str | None:
		return None if self.hitsound == 'none' else self.hitsound.replace('_', '.')
	

	def iter_arctaps(self) -> Iterator[Arctap]:
		yield from self._arctaps
	
	@property
	def arctaps(self) -> tuple[Arctap, ...]:
		return tuple(self._arctaps)
	
	def find_arctap(self, arctap: Arctap) -> int:
		for i, at in enumerate(self._arctaps):
			if at is arctap: # check identity (`is`) because we are maintaining refs
				return i
		return -1
	
	def add_arctap(self, arctap: Arctap) -> None:
		if arctap in self:
			raise ValueError('Arctap is already added to this arc')
		arctap.bind_to(self)
	
	def remove_arctap(self, arctap: Arctap) -> None:
		if arctap not in self:
			raise ValueError('Arctap is not found in this arc')
		arctap.unbind()
	
	def clear_arctaps(self) -> None:
		for at in self._arctaps:
			at.unbind()
	
	def extend_arctaps(self, arctaps: Iterable[Arctap]) -> None:
		for at in arctaps:
			self.add_arctap(at)

	def _get_extras(self) -> Any:
		return self._arctaps
	
	def _set_extras(self, extras: Any) -> None:
		try:
			arctaps = list(map(Arctap.from_any, list(extras)))
			for arctap in arctaps:
				arctap.bind_to(self)
		except (ValueError, TypeError) as e:
			raise ValueError(e)
	
	def position_at(self, time: int) -> tuple[float, float]:
		x = self.easing_x.at(time)
		y = self.easing_y.at(time)
		return (x, y)

	def __contains__(self, arctap: Arctap) -> bool:
		return self.find_arctap(arctap) != -1
	
	@classmethod
	def from_scaled_arctap(cls, sat: ScaledArctap) -> Self:
		return cls(
			begin_time = sat.time,
			end_time = sat.time,
			begin_x = sat.begin_x,
			end_x = sat.end_x,
			easing = ArcEasing.S,
			begin_y = sat.y,
			end_y = sat.y,
			color = ArcColor.Grey,
			hitsound = sat.hitsound,
			type_ = ArcType.Solid,
		)
	
	def to_scaled_arctap(self) -> ScaledArctap:
		return ScaledArctap.from_arc(self)
	
	def is_scaled_arctap(self) -> bool:
		try:
			self.to_scaled_arctap()
			return True
		except ValueError:
			return False

	def _after_validation(self) -> Self:
		super()._after_validation()

		if all([
			self.begin_time == self.end_time,
			self.begin_x == self.end_x,
			self.begin_y == self.end_y
		]):
			raise ValueError(f'The start and end points of arc cannot be the same')

		return self
	
	def clip(
		self,
		begin: uint,
		end: uint,
		/, *, 
		preserve_arctaps: bool = True
	) -> Arc:
		
		begin = int(begin)
		end = int(end)
		
		arc = self.model_copy()
		arc.time_range = begin, end

		if any([
			begin < self.begin_time,
			end > self.end_time
		]):
			raise ValueError(f'Clip is out of the range of the arc')
		
		begin_x, begin_y = self.position_at(begin)
		end_x, end_y = self.position_at(end)

		arc.clear_arctaps()
		arc.begin_x, arc.begin_y = ArcCoord(begin_x), ArcCoord(begin_y)
		arc.end_x, arc.end_y = ArcCoord(end_x), ArcCoord(end_y)

		if preserve_arctaps:
			arc.extend_arctaps(
				at for at in self._arctaps
				if at.time >= begin and at.time <= end
			)

		return arc
	
	def copy(self) -> Self:
		return self.__copy__()
	
	def deepcopy(self) -> Self:
		return self.__deepcopy__()
	
	def __copy__(self) -> Self:
		cls = self.__class__
		data = {
			name: getattr(self, name)
			for name in cls.fields_order
		}
		return cls(**data)
	
	def __deepcopy__(self, memo: dict[int, Any] | None = None) -> Self:
		instance = self.__copy__()
		for at in self.arctaps:
			copied_at = at.__deepcopy__()
			copied_at.bind_to(instance)

		return instance
	
	def mirror(self) -> None:
		self.begin_x = ArcCoord(1.0 - self.begin_x)
		self.end_x = ArcCoord(1.0 - self.end_x)


class ScaledArctap(SkyEvent, TapLikeEvent):

	__aff_config__: ClassVar[AFFEventConfig] = AFFEventConfig(
		is_event=True,
		uses_extra_args=True,
		commands=('arc', ),
		argument_order=('time', 'begin_x', 'end_x', 'y', 'hitsound')
	)
	
	time: uint
	begin_x: ArcCoord
	end_x: ArcCoord
	y: ArcCoord
	hitsound: str

	@classmethod
	def from_arc(cls, arc: Arc) -> Self:
		if not all([
			arc.begin_time == arc.end_time,
			arc.begin_y == arc.end_y,
			arc.color == ArcColor.Grey,
			arc.type_ == ArcType.Solid,
			len(arc.arctaps) == 0
		]):
			raise ValueError(f'Failed to convert this arc to a scaled arctap')
		
		return cls(
			time = arc.begin_time,
			begin_x = arc.begin_x,
			end_x = arc.end_x,
			y = arc.begin_y,
			hitsound = arc.hitsound
		)

	def to_arc(self) -> Arc:
		return Arc.from_scaled_arctap(self)
	
	@classmethod
	def from_str(cls, line: str) -> Self:
		arc = Arc.from_str(line)
		return cls.from_arc(arc)

	def to_str(self) -> str:
		return self.to_arc().to_str()
	

	@classmethod
	def from_dict(cls, data: dict) -> Self:
		arc = Arc.from_dict(data)
		return cls.from_arc(arc)
	
	@classmethod
	def from_analysed(cls, command: str, args: tuple, extras: tuple | None = None) -> Self:
		arc = Arc.from_analysed(command, args, extras)
		return cls.from_arc(arc)
	
	@classmethod
	def from_args(cls, args: tuple, extras: tuple | None = None) -> Self:
		arc = Arc.from_args(args, extras)
		return cls.from_arc(arc)
	
	def mirror(self) -> None:
		self.begin_x = Coordinate.mirror_arc(self.begin_x)
		self.end_x = Coordinate.mirror_arc(self.end_x)



from __future__ import annotations

from math import isfinite
from re import fullmatch
from typing import Any, Self, final

from pydantic_core.core_schema import CoreSchema, float_schema, no_info_plain_validator_function, plain_serializer_function_ser_schema

from mortis.utils import classproperty



__all__ = ['dffloat', 'dffloat2', 'dffloat3']

class _dffloat_base(float):
	"""
	Inner type implementing `dffloat`.
	"""

	__pattern__ = r'^[+-]?(?:\d+\.\d*|\.\d+)$'
	__slots__ = ()

	@classproperty
	def precision(cls) -> int | None:
		raise NotImplementedError
	
	@final
	@classproperty
	def pattern(cls) -> str:
		return cls.__pattern__

	def __new__(cls, value: Any) -> Self:
		if isinstance(value, str):
			value = value.strip()
			if not fullmatch(cls.pattern, value):
				raise ValueError(f'Value {value!r} does not match with the required pattern {cls.pattern!r}')
		
		vf = float(value)
		if not isfinite(vf):
			raise ValueError(f'Value must be finite, got {vf}')
		
		if cls.precision is not None: # default behaviour: no rounding
			vf = round(vf, cls.precision)

		instance = super().__new__(cls, vf)
		return instance
		
	def __str__(self) -> str:
		vf = float(self)

		if self.precision is not None:
			value_str = f'{vf:.{self.precision}f}'
		else:
			value_str = str(vf)
		
		if '.' not in value_str:
			return value_str + '.0'
		
		return value_str
	
	def __repr__(self) -> str:
		return f'{self.__class__.__name__}({str(self)})'
	
	@final
	@classmethod
	def __get_pydantic_core_schema__(cls, source_type, handler) -> CoreSchema:

		def validate(value):
			if isinstance(value, cls):
				return value
			try:
				return cls(value)
			except TypeError as e:
				raise ValueError(e) from e

		return no_info_plain_validator_function(
			validate,
			serialization=plain_serializer_function_ser_schema(
				lambda v: float(v),
				return_schema=float_schema()
			)
		)


__mortis_dffloats__: dict[int, type[_dffloat_base]] = {}
class dffloat(_dffloat_base):
	"""
	Subclass of `float` for ensuring format requirement of AFF file. 

		value_2 = dffloat(3.0)
		# Both declarations are equivalent to each other.

	This class behaves like normal `float` in most cases, but:
	- Arithmetics, etc. applied on instances, would returns a result in `float`, not `dffloat`.
	  - This is because `dffloat` failed to maintain (mathematical) closure from `float`.
	    - Or simpler, no way to ensure results of `dffloat` operations is still a valid `dffloat`.
	  - This always happens, **even both operands are `dffloat`.**:

		value_3 = value_1 + value_2

		print(type(value_3))
		>>> <class 'float'>
		print(value_3)
		>>> 6.0

	- This rejects nonfinite values e.g. `inf` and `nan`.:

		# dffloat('nan')
		# This would result in a ValueError

	- When parsing from a `str`, a `[+/-][aaa].[bbb]` format `str` is required, where
	  - `[...]` indicates that `...` is optional, and
	  - at least one of `[aaa]` and `[bbb]` should be provided.:

		print(dffloat('1.0'))
		>>> 1.0

		print(dffloat('+7.'))
		>>> 7.0

		print(dffloat('-.8'))
		>>> -0.8

		# dffloat('1')
		# ValueError, since '1' does not match the pattern. There is no decimal point in '1'.
	
	- You can limit the float precision (see below).

	These requirements sounds totally weird and counter-intuitive, but AFF file format **does need such behaviours**.
	
	Using `dffloat[x]` can create a **new type** which limits the maximum float precision to posint `x`.:

		value_5 = dffloat[3](1.0)

		print(value_5)
		>>> 1.000

		type(value_5)
		# dffloat[3]

	- Any of `dffloat[x]` is **not** a subclass of `dffloat`.:

		print(issubclass(dffloat[3], dffloat))
		>>> False

	- Using `dffloat` without specifying `[x]` is ok as well.
	  - In this case, there is no float precision limitations.:
	
		print(dffloat(1.24))
		>>> 1.24
		print(dffloat(1.24444))
		>>> 1.24444
	
	- Use `cls.precision` to check its float precision.:

		print(dffloat.precision)
		>>> None
		print(dffloat[3].precision)
		>>> 3

	"""
	@classproperty
	def precision(cls) -> int | None:
		return None

	def __class_getitem__(cls, prec: int) -> type[_dffloat_base]:
		"""
		Syntactic sugar for creating `dffloat[x]` classes with precision limitations.
		"""
		if not isinstance(prec, int):
			raise TypeError(f'Precision must be of type int, got value {prec!r} of type {type(prec)!r}')
		
		if prec <= 0:
			raise TypeError(f'Precision must be a positive int, got {prec}')
		
		if prec not in __mortis_dffloats__:
			class Subclass(_dffloat_base):
				@classproperty
				def precision(cls) -> int | None:
					return prec
				
			Subclass.__name__ = f'{dffloat.__name__}[{prec}]'
			Subclass.__qualname__ = f'{dffloat.__qualname__}[{prec}]'
			
			__mortis_dffloats__[prec] = Subclass
		
		return __mortis_dffloats__[prec]
	
dffloat2 = dffloat[2]
dffloat3 = dffloat[3]


# :: AUTOGEN-START ::

from .aff import AFF
from .events import ArcType, ArcColor, Arctap, Arc, ScaledArctap, CameraEasing, Camera, Hold, HoldFloat, SceneControlType, SceneControl, Tap, TapFloat, Timing
from .lexer import analyse_annotation, analyse_part_separation, analyse_timinggroup_header, analyse_timinggroup_footer, analyse_command, TokenType, Token, Tokens, ARG_TOKEN_TYPES, COMMAND_TOKEN_TYPES, TOKEN_TYPES_PATTERNS, tokenize
from .timinggroup import parse_event, TimingGroup
from .types import FixedLane, FloatLane, ArcCoord, Coordinate, dffloat, dffloat2, dffloat3, BaseEasing, EasingLinear, EasingSineIn, EasingSineOut, EasingBezierDefault, EasingSineInOut, ArcEasing, get_easing_x, get_easing_y, AFFEventConfig, AFFEvent, GameObjectEvent, TechnicalEvent, FloorEvent, SkyEvent, LongNoteEvent, TapLikeEvent

__all__ = [
    'AFF', 
    'ArcType', 
    'ArcColor', 
    'Arctap', 
    'Arc', 
    'ScaledArctap', 
    'CameraEasing', 
    'Camera', 
    'Hold', 
    'HoldFloat', 
    'SceneControlType', 
    'SceneControl', 
    'Tap', 
    'TapFloat', 
    'Timing', 
    'analyse_annotation', 
    'analyse_part_separation', 
    'analyse_timinggroup_header', 
    'analyse_timinggroup_footer', 
    'analyse_command', 
    'TokenType', 
    'Token', 
    'Tokens', 
    'ARG_TOKEN_TYPES', 
    'COMMAND_TOKEN_TYPES', 
    'TOKEN_TYPES_PATTERNS', 
    'tokenize', 
    'parse_event', 
    'TimingGroup', 
    'FixedLane', 
    'FloatLane', 
    'ArcCoord', 
    'Coordinate', 
    'dffloat', 
    'dffloat2', 
    'dffloat3', 
    'BaseEasing', 
    'EasingLinear', 
    'EasingSineIn', 
    'EasingSineOut', 
    'EasingBezierDefault', 
    'EasingSineInOut', 
    'ArcEasing', 
    'get_easing_x', 
    'get_easing_y', 
    'AFFEventConfig', 
    'AFFEvent', 
    'GameObjectEvent', 
    'TechnicalEvent', 
    'FloorEvent', 
    'SkyEvent', 
    'LongNoteEvent', 
    'TapLikeEvent', 
]

# :: AUTOGEN-END ::



from __future__ import annotations

from typing import ClassVar

from pydantic import NonNegativeInt as uint

from mortis.aff.types.models import AFFEventConfig, FloorEvent, LongNoteEvent
from mortis.aff.types.coord import Coordinate, FixedLane, FloatLane


__all__ = ['Hold', 'HoldFloat']

class Hold(FloorEvent, LongNoteEvent):

	__aff_config__: ClassVar[AFFEventConfig] = AFFEventConfig(
		is_event=True,
		commands=('hold', ),
		argument_order=('begin_time', 'end_time', 'lane')
	)

	begin_time: uint
	end_time: uint
	lane: FixedLane

	def to_fixed_lane(self) -> Hold:
		return self

	def to_float_lane(self) -> HoldFloat:
		return HoldFloat(
			begin_time = self.begin_time,
			end_time = self.end_time,
			lane = Coordinate.fixed_to_float(self.lane),
		)
	
	def mirror(self) -> None:
		self.lane = Coordinate.mirror_fixed(self.lane)


class HoldFloat(Hold):

	lane: FloatLane

	def to_float_lane(self) -> HoldFloat:
		return self

	def to_fixed_lane(self) -> Hold:
		return Hold(
			begin_time = self.begin_time,
			end_time = self.end_time,
			lane = Coordinate.float_to_fixed(self.lane),
		)
	
	def mirror(self) -> None:
		self.lane = Coordinate.mirror_float(self.lane) # type: ignore


# :: AUTOGEN-START ::

from .arc import ArcType, ArcColor, Arctap, Arc, ScaledArctap
from .camera import CameraEasing, Camera
from .hold import Hold, HoldFloat
from .scenecontrol import SceneControlType, SceneControl
from .tap import Tap, TapFloat
from .timing import Timing

__all__ = [
    'ArcType', 
    'ArcColor', 
    'Arctap', 
    'Arc', 
    'ScaledArctap', 
    'CameraEasing', 
    'Camera', 
    'Hold', 
    'HoldFloat', 
    'SceneControlType', 
    'SceneControl', 
    'Tap', 
    'TapFloat', 
    'Timing', 
]

# :: AUTOGEN-END ::



from __future__ import annotations

from enum import IntEnum

from mortis.aff.types.dffloat import dffloat2, dffloat3


__all__ = ['FixedLane', 'FloatLane', 'ArcCoord', 'Coordinate']

class FixedLane(IntEnum):
	
	Leftmost = Leftmost6k = Lane0 = 0
	Leftmost4k = Lane1 = 1
	Lane2 = 2
	Lane3 = 3
	Rightmost4k = Lane4 = 4
	Rightmost = Rightmost6k = Lane5 = 5

ArcCoord = dffloat2
FloatLane = dffloat3

class Coordinate:
	@staticmethod
	def arc_to_fixed(value: ArcCoord) -> FixedLane:
		arc = ArcCoord(value)
		return (
			FixedLane.Leftmost
			if (raw := round(ArcCoord(arc) * 2 + 1.5)) < FixedLane.Leftmost
			else (
				FixedLane.Rightmost
				if raw > FixedLane.Rightmost
				else FixedLane(raw)
			)
		)
	
	@staticmethod
	def arc_to_float(value: ArcCoord) -> FloatLane:
		arc = ArcCoord(value)
		return FloatLane(arc / 2 + 0.25)

	@staticmethod
	def fixed_to_arc(value: FixedLane) -> ArcCoord:
		fixed_lane = FixedLane(value)
		return ArcCoord(fixed_lane.value / 2 - 0.75)

	@staticmethod
	def fixed_to_float(value: FixedLane) -> FloatLane:
		fixed_lane = FixedLane(value)
		return FloatLane(fixed_lane.value / 4 - 0.125)
	
	@staticmethod
	def float_to_arc(value: FloatLane) -> ArcCoord:
		float_lane = FloatLane(value)
		return ArcCoord(float_lane * 2 - 0.5)
	
	@staticmethod
	def float_to_fixed(value: FloatLane) -> FixedLane:
		float_lane = FloatLane(value)
		return (
			FixedLane.Leftmost
			if (raw := round(float_lane * 4 + 0.5)) < FixedLane.Leftmost
			else (
				FixedLane.Rightmost
				if raw > FixedLane.Rightmost
				else FixedLane(raw)
			)
		)
	
	@staticmethod
	def mirror_arc(value: ArcCoord) -> ArcCoord:
		return ArcCoord(1.0 - ArcCoord(value))
	
	@staticmethod
	def mirror_float(value: FloatLane) -> FloatLane:
		return FloatLane(1.0 - FloatLane(value))
	
	@staticmethod
	def mirror_fixed(value: FixedLane) -> FixedLane:
		return FixedLane(5 - value.value)